@@ -8,11 +8,7 @@
 #include "pgrn-value.h"
 
 #include <access/heapam.h>
-#ifdef PGRN_SUPPORT_TABLEAM
-#	include <access/tableam.h>
-#else
-#	include <access/htup_details.h>
-#endif
+#include <access/tableam.h>
 #include <access/xact.h>
 #include <access/xlog.h>
 #include <catalog/pg_database.h>
@@ -420,8 +416,10 @@ pgroonga_crash_safer_flush_one(Datum databaseInfoDatum)
 
 	GRN_LOG(&ctx,
 			GRN_LOG_NOTICE,
-			TAG ": initialize: <%s>",
-			PGRN_VERSION);
+			TAG ": initialize: <%s>: %u/%u",
+			PGRN_VERSION,
+			databaseOid,
+			tableSpaceOid);
 
 	grn_ctx_set_wal_role(&ctx, GRN_WAL_ROLE_PRIMARY);
 
@@ -543,16 +541,77 @@ pgroonga_crash_safer_flush_one(Datum databaseInfoDatum)
 
 		/* TODO: How to implement safe finish on no connection? */
 		/*
-		if (pgrn_crash_safer_statuses_get_n_using_processing(statuses,
-															 databaseOid,
-															 tableSpaceOid) == 0)
+		if (pgrn_crash_safer_statuses_get_n_using_processes(statuses,
+															databaseOid,
+															tableSpaceOid) == 0)
 			break;
 		*/
 
 		grn_obj_flush_recursive(&ctx, db);
 	}
 
+	PGroongaCrashSaferGotSIGTERM = false;
+	while (!PGroongaCrashSaferGotSIGTERM && pgrn_file_exist(pgrnDatabasePath))
+	{
+		int conditions;
+		uint32 n_using_processes =
+			pgrn_crash_safer_statuses_get_n_using_processes(statuses,
+															databaseOid,
+															tableSpaceOid);
+		if (n_using_processes == 0)
+			break;
+
+		GRN_LOG(&ctx,
+				GRN_LOG_NOTICE,
+				TAG ": waiting for connections to finish: %u: %u/%u",
+				n_using_processes,
+				databaseOid,
+				tableSpaceOid);
+		conditions = WaitLatch(MyLatch,
+							   WL_LATCH_SET |
+							   WL_TIMEOUT |
+							   PGRN_WL_EXIT_ON_PM_DEATH,
+							   PGroongaCrashSaferFlushNaptime * 1000,
+							   PG_WAIT_EXTENSION);
+		if (conditions & WL_LATCH_SET)
+		{
+			ResetLatch(MyLatch);
+			CHECK_FOR_INTERRUPTS();
+		}
+
+		if (PGroongaCrashSaferGotSIGHUP)
+		{
+			PGroongaCrashSaferGotSIGHUP = false;
+			ProcessConfigFile(PGC_SIGHUP);
+		}
+	}
+
+	if (pgrn_file_exist(pgrnDatabasePath))
+	{
+		GRN_LOG(&ctx,
+				GRN_LOG_NOTICE,
+				TAG ": flushing database before closing: %u/%u",
+				databaseOid,
+				tableSpaceOid);
+		grn_obj_flush_recursive(&ctx, db);
+		GRN_LOG(&ctx,
+				GRN_LOG_NOTICE,
+				TAG ": flushed database before closing: %u/%u",
+				databaseOid,
+				tableSpaceOid);
+	}
+
+	GRN_LOG(&ctx,
+			GRN_LOG_NOTICE,
+			TAG ": closing database: %u/%u",
+			databaseOid,
+			tableSpaceOid);
 	grn_obj_close(&ctx, db);
+	GRN_LOG(&ctx,
+			GRN_LOG_NOTICE,
+			TAG ": closed database: %u/%u",
+			databaseOid,
+			tableSpaceOid);
 
 	grn_ctx_fin(&ctx);
 
@@ -607,7 +666,7 @@ pgroonga_crash_safer_main_flush_all(void)
 	HTAB *statuses;
 	const LOCKMODE lock = AccessShareLock;
 	Relation pg_database;
-	PGrnTableScanDesc scan;
+	TableScanDesc scan;
 	HeapTuple tuple;
 
 	StartTransactionCommand();
@@ -617,8 +676,8 @@ pgroonga_crash_safer_main_flush_all(void)
 
 	statuses = pgrn_crash_safer_statuses_get();
 
-	pg_database = pgrn_table_open(DatabaseRelationId, lock);
-	scan = pgrn_table_beginscan_catalog(pg_database, 0, NULL);
+	pg_database = table_open(DatabaseRelationId, lock);
+	scan = table_beginscan_catalog(pg_database, 0, NULL);
 	for (tuple = heap_getnext(scan, ForwardScanDirection);
 		 HeapTupleIsValid(tuple);
 		 tuple = heap_getnext(scan, ForwardScanDirection))
@@ -630,11 +689,7 @@ pgroonga_crash_safer_main_flush_all(void)
 		char pgrnDatabasePath[MAXPGPATH];
 		pgrn_crash_safer_statuses_entry *entry;
 
-#ifdef PGRN_FORM_PG_DATABASE_HAVE_OID
 		databaseOid = form->oid;
-#else
-		databaseOid = HeapTupleGetOid(tuple);
-#endif
 		tableSpaceOid = form->dattablespace;
 
 		databasePath = GetDatabasePath(databaseOid, tableSpaceOid);
@@ -651,8 +706,8 @@ pgroonga_crash_safer_main_flush_all(void)
 												 NULL);
 		pgroonga_crash_safer_main_flush_one(entry);
 	}
-	pgrn_table_endscan(scan);
-	pgrn_table_close(pg_database, lock);
+	table_endscan(scan);
+	table_close(pg_database, lock);
 
 	PopActiveSnapshot();
 	CommitTransactionCommand();
